// Camera module for ESP-CAM
// Partly based on: // https://gist.github.com/me-no-dev/d34fba51a8f059ac559bf62002e61aa3
//
#include <ESPAsyncWiFiManager.h>  // https://github.com/alanswx/ESPAsyncWiFiManager
#include <esp_camera.h>
#include <SPIFFS.h>
//#include "ArduinoJson.h"

#include "cam.h"
#include "Main.h"
#include "Webserver.h"
#include "cfg.h"
#include "hw.h"

// Define private used funcs
//bool GetJsonField_UpdateCam(sensor_t *s, DynamicJsonDocument doc, char *variable);

typedef struct
{
    camera_fb_t *fb;
    size_t index;
} camera_frame_t;

#define PART_BOUNDARY "123456789000000000000987654321"
static const char *STREAM_CONTENT_TYPE = "multipart/x-mixed-replace;boundary=" PART_BOUNDARY;
static const char *STREAM_BOUNDARY = "\r\n--" PART_BOUNDARY "\r\n";
static const char *STREAM_PART = "Content-Type: %s\r\nContent-Length: %u\r\n\r\n";

static const char *JPG_CONTENT_TYPE = "image/jpeg";
//static const char *BMP_CONTENT_TYPE = "image/x-windows-bmp";

class AsyncBufferResponse : public AsyncAbstractResponse {
   private:
    uint8_t *_buf;
    size_t _len;
    size_t _index;

   public:
    AsyncBufferResponse(uint8_t *buf, size_t len, const char *contentType) {
        _buf = buf;
        _len = len;
        _callback = nullptr;
        _code = 200;
        _contentLength = _len;
        _contentType = contentType;
        _index = 0;
    }
    ~AsyncBufferResponse() {
        if (_buf != nullptr) {
            free(_buf);
        }
    }
    bool _sourceValid() const { return _buf != nullptr; }
    virtual size_t _fillBuffer(uint8_t *buf, size_t maxLen) override {
        size_t ret = _content(buf, maxLen, _index);
        if (ret != RESPONSE_TRY_AGAIN) {
            _index += ret;
        }
        return ret;
    }
    size_t _content(uint8_t *buffer, size_t maxLen, size_t index) {
        memcpy(buffer, _buf + index, maxLen);
        if ((index + maxLen) == _len) {
            free(_buf);
            _buf = nullptr;
        }
        return maxLen;
    }
};

class AsyncFrameResponse : public AsyncAbstractResponse {
   private:
    camera_fb_t *fb;
    size_t _index;

   public:
    AsyncFrameResponse(camera_fb_t *frame, const char *contentType) {
        _callback = nullptr;
        _code = 200;
        _contentLength = frame->len;
        _contentType = contentType;
        _index = 0;
        fb = frame;
    }
    ~AsyncFrameResponse() {
        if (fb != nullptr) {
            esp_camera_fb_return(fb);
        }
    }
    bool _sourceValid() const { return fb != nullptr; }
    virtual size_t _fillBuffer(uint8_t *buf, size_t maxLen) override {
        size_t ret = _content(buf, maxLen, _index);
        if (ret != RESPONSE_TRY_AGAIN) {
            _index += ret;
        }
        return ret;
    }
    size_t _content(uint8_t *buffer, size_t maxLen, size_t index) {
        memcpy(buffer, fb->buf + index, maxLen);
        if ((index + maxLen) == fb->len) {
            esp_camera_fb_return(fb);
            fb = nullptr;
        }
        return maxLen;
    }
};

class AsyncJpegStreamResponse : public AsyncAbstractResponse {
   private:
    camera_frame_t _frame;
    size_t _index;
    size_t _jpg_buf_len;
    uint8_t *_jpg_buf;
//    long lastAsyncRequest;

   public:
    AsyncJpegStreamResponse() {
        _callback = nullptr;
        _code = 200;
        _contentLength = 0;
        _contentType = STREAM_CONTENT_TYPE;
        _sendContentLength = false;
        _chunked = true;
        _index = 0;
        _jpg_buf_len = 0;
        _jpg_buf = NULL;
//        lastAsyncRequest = 0;
        memset(&_frame, 0, sizeof(camera_frame_t));
    }
    ~AsyncJpegStreamResponse() {
        if (_frame.fb) {
            if (_frame.fb->format != PIXFORMAT_JPEG) {
                free(_jpg_buf);
            }
            esp_camera_fb_return(_frame.fb);
        }
    }
    bool _sourceValid() const {
        return true;
    }
    virtual size_t _fillBuffer(uint8_t *buf, size_t maxLen) override {
        size_t ret = _content(buf, maxLen, _index);
        if (ret != RESPONSE_TRY_AGAIN) {
            _index += ret;
        }
        return ret;
    }
    size_t _content(uint8_t *buffer, size_t maxLen, size_t index) {
        if (!_frame.fb || _frame.index == _jpg_buf_len) {
            if (index && _frame.fb) {
//                long end = millis();
//                int fp = (end - lastAsyncRequest);
//                Log.printf("Size: %uKB, Time: %ums (%ifps)\n", _jpg_buf_len / 1024, fp, 1000 / fp);
//                lastAsyncRequest = end;
                if (_frame.fb->format != PIXFORMAT_JPEG) {
                    free(_jpg_buf);
                }
                esp_camera_fb_return(_frame.fb);
                _frame.fb = NULL;
                _jpg_buf_len = 0;
                _jpg_buf = NULL;
            }
            if (maxLen < (strlen(STREAM_BOUNDARY) + strlen(STREAM_PART) + strlen(JPG_CONTENT_TYPE) + 8)) {
                //Log.printf("Not enough space for headers");
                return RESPONSE_TRY_AGAIN;
            }
            //get frame
            _frame.index = 0;

            _frame.fb = esp_camera_fb_get();
            if (_frame.fb == NULL) {
                Log.printf("Camera frame failed");
                hw_do_vflash(false);
                return 0;
            }

            if (_frame.fb->format != PIXFORMAT_JPEG) {
                unsigned long st = millis();
                bool jpeg_converted = frame2jpg(_frame.fb, 80, &_jpg_buf, &_jpg_buf_len);
                if (!jpeg_converted) {
                    Log.printf("JPEG compression failed");
                    esp_camera_fb_return(_frame.fb);
                    _frame.fb = NULL;
                    _jpg_buf_len = 0;
                    _jpg_buf = NULL;
                    hw_do_vflash(false);
                    return 0;
                }
                Log.printf("JPEG: %lums, %uB", millis() - st, _jpg_buf_len);
            } else {
                _jpg_buf_len = _frame.fb->len;
                _jpg_buf = _frame.fb->buf;
            }

            //send boundary
            size_t blen = 0;
            if (index) {
                blen = strlen(STREAM_BOUNDARY);
                memcpy(buffer, STREAM_BOUNDARY, blen);
                buffer += blen;
            }
            //send header
            size_t hlen = sprintf((char *)buffer, STREAM_PART, JPG_CONTENT_TYPE, _jpg_buf_len);
            buffer += hlen;
            //send frame
            hlen = maxLen - hlen - blen;
            if (hlen > _jpg_buf_len) {
                maxLen -= hlen - _jpg_buf_len;
                hlen = _jpg_buf_len;
            }
            memcpy(buffer, _jpg_buf, hlen);
            _frame.index += hlen;
            return maxLen;
        }

        size_t available = _jpg_buf_len - _frame.index;
        if (maxLen > available) {
            maxLen = available;
        }
        memcpy(buffer, _jpg_buf + _frame.index, maxLen);
        _frame.index += maxLen;

        return maxLen;
    }
};

// ----------------------------------------------------------------------------------------------------------------
// ------  Camera Webfunctions ------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void sendJpg(AsyncWebServerRequest *request) {
    if (!_webAuth_var(request))
        return;
    hw_do_flash(true);
    camera_fb_t *fb = esp_camera_fb_get();
    if (fb == NULL) {
        Log.println(F("Webrequest: \"/capture\" -> Camera not Detected.\n"));
        String s = F("Camera not Detected.<script>setTimeout(function() {window.parent.location.href= \"/\";s}, 3000);</script>");
        request->send(200, F("text/html"), s);
        hw_do_flash(false);
        return;
    }
    Log.println(F("Start JPG Capture\n"));
    if (fb->format == PIXFORMAT_JPEG) {
        AsyncFrameResponse *response = new AsyncFrameResponse(fb, JPG_CONTENT_TYPE);
        if (response == NULL) {
            Log.println(F("Response alloc failed"));
            request->send(501);
            hw_do_flash(false);
            return;
        }
        response->addHeader(F("Access-Control-Allow-Origin"), "*");
        request->send(response);
        hw_do_flash(false);
        return;
    }

    size_t jpg_buf_len = 0;
    uint8_t *jpg_buf = NULL;
    unsigned long st = millis();
    bool jpeg_converted = frame2jpg(fb, 80, &jpg_buf, &jpg_buf_len);
    esp_camera_fb_return(fb);
    if (!jpeg_converted) {
        Log.printf("JPEG compression failed: %lu", millis());
        request->send(501);
        hw_do_flash(false);
        return;
    }
    Log.printf("JPEG: %lums, %uB", millis() - st, jpg_buf_len);

    AsyncBufferResponse *response = new AsyncBufferResponse(jpg_buf, jpg_buf_len, JPG_CONTENT_TYPE);
    if (response == NULL) {
        Log.println(F("Response alloc failed"));
        request->send(501);
        hw_do_flash(false);
        return;
    }
    response->addHeader(F("Access-Control-Allow-Origin"), "*");
    request->send(response);
    hw_do_flash(false);
}

void streamJpg(AsyncWebServerRequest *request) {
    if (!_webAuth_var(request))
        return;
    hw_do_vflash(true);
    camera_fb_t *fb = esp_camera_fb_get();
    if (fb == NULL) {
        Log.println(F("Webrequest: \"/stream\" -> Camera not Detected.\n"));
        String s = F("Camera not Detected.<script>setTimeout(function() {window.parent.location.href= \"/\";s}, 3000);</script>");
        request->send(200, "text/html", s);
        hw_do_vflash(false);
        return;
    }
    Log.println(F("Start JPG streaming\n"));
    AsyncJpegStreamResponse *response = new AsyncJpegStreamResponse();
    if (!response) {
        request->send(501);
        hw_do_vflash(false);
        return;
    }
    response->addHeader("Access-Control-Allow-Origin", "*");
    request->send(response);
}

// ----------------------------------------------------------------------------------------------------------------
// ------  Camera other functions  ------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

// Initialise cameras
bool initcamera() {
    camera_config_t config;

    config.ledc_channel = LEDC_CHANNEL_0;
    config.ledc_timer = LEDC_TIMER_0;
    config.pin_d0 = Y2_GPIO_NUM;
    config.pin_d1 = Y3_GPIO_NUM;
    config.pin_d2 = Y4_GPIO_NUM;
    config.pin_d3 = Y5_GPIO_NUM;
    config.pin_d4 = Y6_GPIO_NUM;
    config.pin_d5 = Y7_GPIO_NUM;
    config.pin_d6 = Y8_GPIO_NUM;
    config.pin_d7 = Y9_GPIO_NUM;
    config.pin_xclk = XCLK_GPIO_NUM;
    config.pin_pclk = PCLK_GPIO_NUM;
    config.pin_vsync = VSYNC_GPIO_NUM;
    config.pin_href = HREF_GPIO_NUM;
    config.pin_sscb_sda = SIOD_GPIO_NUM;
    config.pin_sscb_scl = SIOC_GPIO_NUM;
    config.pin_pwdn = PWDN_GPIO_NUM;
    config.pin_reset = RESET_GPIO_NUM;
    config.xclk_freq_hz = 20000000;
    config.pixel_format = PIXFORMAT_JPEG;

    //Setup size of image
    if (psramFound()) {
        Log.println(F("Camera found psram"));
        config.frame_size = FRAMESIZE_VGA;
//        config.frame_size = FRAMESIZE_SVGA;
        config.jpeg_quality = 10;
        config.fb_count = 1;
    } else {
        Log.println(F("Camera without psram"));
        config.frame_size = FRAMESIZE_VGA;
        config.jpeg_quality = 12;
        config.fb_count = 1;
    }

    // Camera init
    esp_err_t err = esp_camera_init(&config);
    if (err != ESP_OK) {
        Log.println(F("Camera init failed!\n"));
        return false;
    }

    // restore the saved settings from SPIFFS
    //Restore_CamSettings_from_SPIFFS();

    Log.println(F("Camera initialised!\n"));
    return true;
}
