#include <Arduino.h>
#include <Main.h>
#include <session.h>
#include "FS.h"
#include "SPIFFS.h"
#include "mbedtls/md5.h"
#include <libb64/cencode.h>

SessionEntry Sessions[SESSION_MAX];

unsigned char itoa64[] =         /* 0 ... 63 => ascii - 64 */
	"az0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

void to64(char *s, uint32_t v, int n)
{
	while (--n >= 0)
	{
		*s++ = itoa64[v&0x3f];
		v >>= 6;
	}
}

bool session_getMD5(uint8_t * data, uint16_t len, char * output)
{	//33 bytes or more
	mbedtls_md5_context _ctx;
	uint8_t i;
	uint8_t * _buf = (uint8_t*)malloc(16);
	if(_buf == NULL)
		return false;
	memset(_buf, 0x00, 16);
	mbedtls_md5_init(&_ctx);
	mbedtls_md5_starts(&_ctx);
	mbedtls_md5_update(&_ctx, data, len);
	mbedtls_md5_finish(&_ctx, _buf);
	for(i = 0; i < 16; i++)
	{
		sprintf(output + (i * 2), "%02x", _buf[i]);
	}
	free(_buf);
	return true;
}

bool session_b64MD5(uint8_t * data, uint16_t len, char * output)
{	//33 bytes or more
	uint32_t l;
	mbedtls_md5_context _ctx;
	char *p;
	uint8_t * _buf = (uint8_t*)malloc(16);

	if(_buf == NULL)
		return false;
	memset(_buf, 0x00, 16);
	mbedtls_md5_init(&_ctx);
	mbedtls_md5_starts(&_ctx);
	mbedtls_md5_update(&_ctx, data, len);
	mbedtls_md5_finish(&_ctx, _buf);

	p=output;

	l = (_buf[ 0]<<16) | (_buf[ 6]<<8) | _buf[12]; to64(p,l,4); p += 4;
	l = (_buf[ 1]<<16) | (_buf[ 7]<<8) | _buf[13]; to64(p,l,4); p += 4;
	l = (_buf[ 2]<<16) | (_buf[ 8]<<8) | _buf[14]; to64(p,l,4); p += 4;
	l = (_buf[ 3]<<16) | (_buf[ 9]<<8) | _buf[15]; to64(p,l,4); p += 4;
	l = (_buf[ 4]<<16) | (_buf[10]<<8) | _buf[ 5]; to64(p,l,4); p += 4;
	l = _buf[11]; to64(p,l,2); p += 2;
	*p = '\0';

//	base64_encode_chars((char *)_buf, 16, output);
	free(_buf);
	return true;
}

String session_genRandomMD5()
{
	uint32_t r = rand();
	char * out = (char*)malloc(33);
	if(out == NULL || !session_getMD5((uint8_t*)(&r), 4, out))
		return "";
	String res = String(out);
	free(out);
	return res;
}

String session_genRandomb64MD5()
{
	uint32_t r[4];
	r[0] = esp_random();
	r[1] = esp_random();
	r[2] = esp_random();
	r[3] = esp_random();
	char * out = (char*)malloc(33);
	if(out == NULL || !session_b64MD5((uint8_t*)(r), 16, out))
		return "";
	String res = String(out);
	free(out);
	return res;
}

String session_stringMD5(const String& in)
{
	char * out = (char*)malloc(33);
	if(out == NULL || !session_getMD5((uint8_t*)(in.c_str()), in.length(), out))
		return "";
	String res = String(out);
	free(out);
	return res;
}

void session_print(void)
{
	for(uint8_t i=0;i<SESSION_MAX;i++)
	{
		if(Sessions[i].time)
		{
			Log.printf("%02d %d %s\n\r",i,Sessions[i].time,Sessions[i].sid);
		}
	}
}

void session_clear_n(uint8_t n)
{
	if(n>=SESSION_MAX) return;
	Sessions[n].time=0;
	Sessions[n].sid[0]='\0';
}

void session_clear_all(void)
{
	// init sessions
	for(uint8_t i=0;i<SESSION_MAX;i++)
	{
		session_clear_n(i);
	}
}

bool session_read(void)
{
	File file = SPIFFS.open("/esid", "r");
	if (file) 
	{
		size_t len=0;
		len=file.read((uint8_t *)&Sessions,sizeof(Sessions));
		file.close();

		if(len) return true;
	}
	return false;
}

bool session_write(void)
{
	File file = SPIFFS.open("/esid", "w");
	if (file) 
	{
		size_t len=0;
		len=file.write((uint8_t *)&Sessions,sizeof(Sessions));
		file.close();

		if(len) return true;
	}
	return false;
}

void session_touch(uint8_t n)
{
		time_t now;
		time(&now);

		Sessions[n].time=now;
}

int8_t session_search(char *sid)
{
	for(uint8_t i=0;i<SESSION_MAX;i++)
	{
		if(Sessions[i].time != 0 && strcmp(Sessions[i].sid,sid)==0)
		{
			return i;
		}
	}
	return -1;
}

int8_t session_search_touch(char *sid)
{
	int8_t ret=session_search(sid);
	if(ret>=0)
	{
		session_touch(ret);
		session_write();
	}
	return ret;
}

bool session_clear(char *sid)
{
	int8_t ret=session_search(sid);
	if(ret>=0)
	{
		session_clear_n(ret);
		return session_write();
	}
	return false;
}

uint8_t session_new(char *sid)
{
	uint32_t last=UINT32_MAX;
	uint8_t lastn=0;

	// find free slot
	for(uint8_t i=0;i<SESSION_MAX;i++)
	{
		if(Sessions[i].time==0)
		{
			lastn=i;
			break;
		}

		if(Sessions[i].time<last)
		{
			last=Sessions[i].time;
			lastn=i;
		}
	}
	strncpy(Sessions[lastn].sid,sid,SESSION_LEN_MAX-1);
	Sessions[lastn].sid[SESSION_LEN_MAX-1]='\0';
	session_touch(lastn);
	return lastn;
}

uint8_t session_set(char *sid)
{
	int8_t ret=session_search(sid);
Log.printf("Session Set: %s\n\r",sid);
	if(ret>=0)
	{
		session_touch(ret);
	}
	else
	{
		ret = session_new(sid);
	}
	session_write();
	return ret;
}

uint8_t session_gen(void)
{
	return session_set((char *)(session_genRandomb64MD5().c_str()));
//	return session_set((char *)(session_genRandomMD5().c_str()));
}

void session_del(void)
{
	session_clear_all();
	session_write();
}

void session_init(void)
{
	session_clear_all();
	session_read();
}
